import math
import torch
import torch.nn as nn
import torch.utils
import torch.utils.data
#from torchvision import datasets, transforms
from torch.autograd import Variable


"""implementation of the Variational Recurrent
Neural Network (VRNN) from https://arxiv.org/abs/1506.02216
using unimodal isotropic gaussian distributions for 
inference, prior, and generating models."""


class VRNN(nn.Module):
  def __init__(self, x_dim, hlv_dim, h_dim, z_dim, n_layers, train_hlvs, bias=False):
    super(VRNN, self).__init__()

    self.x_dim = x_dim
    self.hlv_dim = hlv_dim
    self.h_dim = h_dim
    self.z_dim = z_dim
    self.n_layers = n_layers
    self.train_hlvs = train_hlvs

    #feature-extracting transformations
    self.phi_x = nn.Sequential(
      nn.Linear(x_dim, h_dim),
      nn.ReLU(),
      nn.Linear(h_dim, h_dim),
      nn.ReLU()).cpu()
    self.phi_z = nn.Sequential(
      nn.Linear(z_dim, h_dim),
      nn.ReLU()).cpu()

    #encoder
    if(self.train_hlvs):
      self.enc = nn.Sequential(
        nn.Linear(h_dim + h_dim + hlv_dim, h_dim),
        nn.ReLU(),
        nn.Linear(h_dim, h_dim),
        nn.ReLU()).cpu()
    else:
      self.enc = nn.Sequential(
        nn.Linear(h_dim + h_dim, h_dim),
        nn.ReLU(),
        nn.Linear(h_dim, h_dim),
        nn.ReLU()).cpu()
    self.enc_mean = nn.Linear(h_dim, z_dim).cpu()
    self.enc_std = nn.Sequential(
      nn.Linear(h_dim, z_dim),
      nn.Softplus()).cpu()

    #prior
    self.prior = nn.Sequential(
      nn.Linear(h_dim, h_dim),
      nn.ReLU()).cpu()
    self.prior_mean = nn.Linear(h_dim, z_dim).cpu()
    self.prior_std = nn.Sequential(
      nn.Linear(h_dim, z_dim),
      nn.Softplus()).cpu()

    #decoder
    if(self.train_hlvs):
      self.dec = nn.Sequential(
        nn.Linear(h_dim + h_dim + hlv_dim, h_dim),
        nn.ReLU(),
        nn.Linear(h_dim, h_dim),
        nn.ReLU()).cpu()
    else:
      self.dec = nn.Sequential(
        nn.Linear(h_dim + h_dim, h_dim),
        nn.ReLU(),
        nn.Linear(h_dim, h_dim),
        nn.ReLU()).cpu()
    #self.dec_std = nn.Sequential(
    #  nn.Linear(h_dim, x_dim),
    #  nn.Softplus()).cpu()
    #self.dec_mean = nn.Sequential(
    #  nn.Linear(h_dim, x_dim),
    #  nn.Tanh()).cpu()
    self.dec_mean = nn.Sequential(
      nn.Linear(h_dim, x_dim)).cpu()
      

    #recurrence
    if(self.train_hlvs):
      self.rnn = nn.GRU(h_dim + h_dim + hlv_dim, h_dim, n_layers, bias).cpu()
    else:
      self.rnn = nn.GRU(h_dim + h_dim, h_dim, n_layers, bias).cpu()
    #self.rnn = nn.GRU(h_dim + h_dim, h_dim, n_layers, bias).cpu()


  def forward(self, x, y, x_avg, kl_weight):

    if x.is_cuda: device='cuda'
    else: device='cpu'


    all_enc_mean, all_enc_std = [], []
    #all_dec_mean, all_dec_std = [], []
    all_dec_mean = []
    all_dec_mean = torch.empty(x.size(0), x.size(1), x.size(2))
    #kld_loss = 0
    kld_loss = torch.zeros([x.size(1)]).to(device)
    nll_loss = 0
    mseloss = torch.nn.MSELoss()
    l1loss = torch.nn.L1Loss()


    #x_avg = torch.mean(x, 1)
  

    h = Variable(torch.zeros(self.n_layers, x.size(1), self.h_dim)).to(device)
    for t in range(x.size(0)):
      
      phi_x_t = self.phi_x(x[t]).to(device)
      #encoder
      if(self.train_hlvs):
        enc_t = self.enc(torch.cat([phi_x_t, h[-1], y], 1)).to(device)
      else:      
        enc_t = self.enc(torch.cat([phi_x_t, h[-1]], 1)).to(device)
      
      enc_mean_t = self.enc_mean(enc_t).to(device)
      enc_std_t = self.enc_std(enc_t).to(device)

      #prior
      prior_t = self.prior(h[-1]).to(device)
      prior_mean_t = self.prior_mean(prior_t).to(device)
      prior_std_t = self.prior_std(prior_t).to(device)

      #sampling and reparameterization
      z_t = self._reparameterized_sample(enc_mean_t, enc_std_t).to(device)
      phi_z_t = self.phi_z(z_t).to(device)

      #decoder
      if(self.train_hlvs):
        dec_t = self.dec(torch.cat([phi_z_t, h[-1], y], 1)).to(device)
      else:
        dec_t = self.dec(torch.cat([phi_z_t, h[-1]], 1)).to(device)
      dec_mean_t = self.dec_mean(dec_t).to(device)


      #recurrence
      if(self.train_hlvs):
        _, h = self.rnn(torch.cat([phi_x_t, phi_z_t, y], 1).unsqueeze(0), h)
      else:
        _, h = self.rnn(torch.cat([phi_x_t, phi_z_t], 1).unsqueeze(0), h)

      #computing losses
      #kld_loss += self._kld_gauss(enc_mean_t, enc_std_t, prior_mean_t, prior_std_t).to(device) 
      #nll_loss += mseloss(dec_mean_t, x[t]).to(device)
      kld_loss += self._kld_gauss(enc_mean_t, enc_std_t, prior_mean_t, prior_std_t).to(device)*x_avg[t]
      #kld_loss += self._kld_gauss(enc_mean_t, enc_std_t, prior_mean_t, prior_std_t).to(device)*x[t][:,0]
      #nll_loss += mseloss(dec_mean_t, x[t]).to(device)*x_avg[t][0]
      #kld_loss += self._kld_gauss(enc_mean_t, enc_std_t, prior_mean_t, prior_std_t).to(device)
      nll_loss += mseloss(dec_mean_t, x[t]).to(device)

      all_enc_std.append(enc_std_t)
      all_enc_mean.append(enc_mean_t)
      all_dec_mean[t] = dec_mean_t.detach()

    kld_loss /= (x.size(0))
    nll_loss /= x.size(0)
    #loss = 0.1*torch.mean(kld_loss) + nll_loss
    loss = kl_weight*torch.mean(kld_loss) + nll_loss


    return kld_loss, nll_loss, loss, all_dec_mean


  def sample(self, seq_len):

    sample = torch.zeros(seq_len, self.x_dim)

    h = Variable(torch.zeros(self.n_layers, 1, self.h_dim))
    for t in range(seq_len):

      #prior
      prior_t = self.prior(h[-1])
      prior_mean_t = self.prior_mean(prior_t)
      prior_std_t = self.prior_std(prior_t)

      #sampling and reparameterization
      z_t = self._reparameterized_sample(prior_mean_t, prior_std_t)
      phi_z_t = self.phi_z(z_t)
      
      #decoder
      dec_t = self.dec(torch.cat([phi_z_t, h[-1]], 1))
      dec_mean_t = self.dec_mean(dec_t)
      #dec_std_t = self.dec_std(dec_t)

      phi_x_t = self.phi_x(dec_mean_t)

      #recurrence
      _, h = self.rnn(torch.cat([phi_x_t, phi_z_t], 1).unsqueeze(0), h)

      sample[t] = dec_mean_t.data
  
    return sample


  def reset_parameters(self, stdv=1e-1):
    for weight in self.parameters():
      weight.data.normal_(0, stdv)


  def _init_weights(self, stdv):
    pass


  def _reparameterized_sample(self, mean, std):
    """using std to sample"""
    eps = torch.FloatTensor(std.size()).normal_().to(mean.device)
    eps = Variable(eps)
    return eps.mul(std).add_(mean).to(mean.device)


  def _kld_gauss(self, mean_1, std_1, mean_2, std_2):
    """Using std to compute KLD"""

    kld_element =  (2 * torch.log(std_2) - 2 * torch.log(std_1) + 
      (std_1.pow(2) + (mean_1 - mean_2).pow(2)) /
      std_2.pow(2) - 1)
    return  0.5 * torch.sum(kld_element, dim=1)


  def _nll_bernoulli(self, theta, x):
    return - torch.sum(x*torch.log(theta) + (1-x)*torch.log(1-theta))


  def _nll_gauss(self, mean, std, x):
    pass

